-- LUA script for use on LabJack T4 with LJTick-InAmp instrumentation amplifier
-- LJTick-InAmp set to gain of 1001

-- initialize variables
local LoopTime = 2 -- ms = 500Hz
local LoadCell = 0 -- force sensor value
local VOut = 0     -- analog output to HV amplifier
local Encoder = 0  -- linear encoder position counts
local oldEncoder = 0  -- prior loop encoder value
local FlagResetEncoder = 0  -- trigger to reset encoder
local Speed = 0    -- calculated linear speed
local NewSpeed = 0 -- difference in position/loop time
local OldSpeed = 0 -- prior calculated speed
local Loops = 0    -- check on completion of control loop in LoopTime
local Error = {}   -- error message return value
local Kp = 0.      -- control law position constant
--local Kp = 160.
local Kv = 0.      -- control law velocity constant 
local ForcePlantar = 0.  -- plantar reaction force set point
local ForceDorsi = 0.    -- dorsi-flextion reaction force set point
local FlagNewParameters = 0  -- trigger to read newly input control parameters

local LED0 = 0      -- LED state 0 for LabJack
local LED1 = 0      -- LED state 1 for LabJack

local Koff = 1.1671 -- voltage offset of load cell/instrumentation amp
local Kcell = 737.5 -- voltage to force constant for load cell/instrumentation amp

local mbWrite=MB.W 
local mbRead=MB.R

LJ.IntervalConfig(0, LoopTime)      --set interval time to LoopTime ms
local checkInterval=LJ.CheckInterval

local i = 0          -- loop counter

MB.writeName("POWER_LED",4)  -- Manual Control of Power LED

-- set up registers for reading encoder
MB.writeName("DIO4_EF_ENABLE",0)
MB.writeName("DIO5_EF_ENABLE",0)
MB.writeName("DIO4_EF_INDEX",10)
MB.writeName("DIO5_EF_INDEX",10)
MB.writeName("DIO4_EF_ENABLE",1)
MB.writeName("DIO5_EF_ENABLE",1)

Error = MB.readNameArray("DIO4_EF_READ_A_AND_RESET", 2, 0) -- reset the encoder counter

-- set LabJack external LEDs to off
Error = MB.writeName("LED_COMM",0)
Error = MB.writeName("LED_STATUS",0)

-- clear input registers for control parameters on startup
MB.writeName("USER_RAM0_U16",0)
MB.writeName("USER_RAM1_U16",0)
MB.writeName("USER_RAM2_U16",0)
MB.writeName("USER_RAM3_U16",0)
MB.writeName("USER_RAM4_U16",0)
MB.writeName("USER_RAM5_U16",0)

-- set LabJack external LEDs to on (until first set of control parameters is received below)
MB.writeName("LED_COMM",1)
MB.writeName("LED_STATUS",1)

while true do -- always do the following loop

  Loops = checkInterval(0) -- check whether loop completes in given interval

  if Loops then   --interval completed

    -- check for updates from Matlab if FlagNewParameters is high
    FlagNewParameters = MB.readName("USER_RAM0_U16")
    if (FlagNewParameters == 1) then
      -- turn off external LEDs to confirm parameters are received
      MB.writeName("LED_COMM",0)
      MB.writeName("LED_STATUS",0)
      ForcePlantar = MB.readName("USER_RAM1_U16") -- read plantar force target
      ForceDorsi = MB.readName("USER_RAM2_U16")   -- read dorsi- force target
      Kp = MB.readName("USER_RAM3_U16")/1000.     -- read and rescale position control constant
      Kv = MB.readName("USER_RAM4_U16")/1000.     -- read and rescale velocity control constant
      FlagResetEncoder = MB.readName("USER_RAM5_U16")  -- read flag to reset encoder (or not)
      -- unset FlagNewParameters so time is not wasted in loop re-reading parameters
      MB.writeName("USER_RAM0_U16",0)
    end

    -- clear encoder if requested
    if true then
      if (FlagResetEncoder == 1) then
        Error = MB.readNameArray("DIO4_EF_READ_A_AND_RESET", 2, 0) -- reset the encoder counter
        oldEncoder = 0 -- clear oldEncoder
        MB.writeName("USER_RAM5_U16",0) -- clear encoder reset flag
        FlagResetEncoder = 0
      end
    end

    -- update oldEncoder and Encoder, calculate speed
    oldEncoder = Encoder
    Encoder = MB.readName("DIO4_EF_READ_A_F") -- 8000 counts/inch = 314.96 counts/mm
    OldSpeed = NewSpeed
    NewSpeed = (Encoder - oldEncoder)/LoopTime -- counts/millisecond
    Speed = 0.5 * (OldSpeed + NewSpeed)  -- simple filter to smooth velocity calculation
    -- if LoadCell < 0, force is in dorsi- direction, if LoadCell > 0, force is plantar
    LoadCell = (MB.readName("AIN3") - Koff) * Kcell / 5. -- read load cell, convert voltage to N

    -- control law calculation
    -- VOut = 0 -- if the system is stationary, set Vout to 0
    -- Speed considered to be zero below 1 count/period
    -- Speed > 0 = dorsi- speed, Speed < 0 = plantar speed
    -- Vout constrained to be between 0 and 4 V

    if (Speed > 0.01) then
      VOut = math.min(math.max(0.,Kp * (ForceDorsi - math.abs(LoadCell))),4.)
    elseif (Speed < -0.01) then
      VOut = math.min(math.max(0.,Kp * (ForcePlantar - math.abs(LoadCell))),4.)
    else
      VOut = 0 -- if the system is stationary, set Vout to 0
    end    

    if (true) then
      MB.writeName("DAC0",VOut) -- VOut in Volts
    end

    i = i + 1 -- increment loop count
    
    -- write output data to internal registers for reading by Matlab
    MB.writeName("USER_RAM0_F32", VOut)
    MB.writeName("USER_RAM1_F32", LoadCell)
    MB.writeName("USER_RAM2_F32", Loops)
    MB.writeName("USER_RAM3_F32", Encoder)
    MB.writeName("USER_RAM4_F32", Speed)
    MB.writeName("USER_RAM5_F32", i)

  end
end
MB.writeName("POWER_LED",1) -- return to normal LED control
MB.writeName("LUA_RUN",0) -- VM unloaded